// $Id$

package org.bolet.jgz.api;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import org.bolet.jgz.ZlibInputStream;
import org.bolet.jgz.ZlibOutputStream;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.util.Arrays;
import java.util.zip.Inflater;

/**
 * Unit tests for the flush modes.
 *
 * @version   $Revision$
 * @author    Thomas Pornin
 */

public class TestFlush extends TestCase {

	public TestFlush(String name)
	{
		super(name);
	}

	public static Test suite()
	{
		return new TestSuite(TestFlush.class);
	}

	/* ============================================================ */

	private void makeFlush(byte[] data1, byte[] data2, int mode)
		throws Exception
	{
		/*
		 * Prepare the compressed data, as two byte arrays (the
		 * flush is performed between the two arrays).
		 */
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		ZlibOutputStream zout = new ZlibOutputStream(baos);
		zout.write(data1);
		switch (mode) {
		case 0:
			zout.flushPartial();
			break;
		case 1:
			zout.flushSync(true);
			break;
		case 2:
			zout.flushFull(true);
			break;
		}
		byte[] zdata1 = baos.toByteArray();
		baos.reset();
		zout.write(data2);
		zout.terminate();
		byte[] zdata2 = baos.toByteArray();

		/*
		 * Test with zlib that all relevant data has been flushed,
		 * and that the inflation may proceed beyond the flush
		 * point.
		 */
		Inflater i = new Inflater();
		baos.reset();
		i.setInput(zdata1);
		byte[] buf = new byte[8192];
		for (;;) {
			int len = i.inflate(buf);
			if (len == 0)
				break;
			baos.write(buf, 0, len);
		}
		if (!i.needsInput())
			fail("inflater does not need input");
		assertTrue(Arrays.equals(data1, baos.toByteArray()));
		baos.reset();
		i.setInput(zdata2);
		for (;;) {
			int len = i.inflate(buf);
			if (len == 0)
				break;
			baos.write(buf, 0, len);
		}
		if (!i.finished())
			fail("inflater not finished");
		assertTrue(Arrays.equals(data2, baos.toByteArray()));

		/*
		 * Test that jgz inflater code handles the flush.
		 */
		ByteArrayInputStream bais = new ByteArrayInputStream(zdata1);
		ZlibInputStream zin = new ZlibInputStream(bais, true);
		baos.reset();
		while (bais.available() >= 2) {
			int len = zin.read(buf);
			if (len < 0)
				fail("unexpected EOF");
			baos.write(buf, 0, len);
		}
		assertTrue(Arrays.equals(data1, baos.toByteArray()));
		baos.reset();
		if (bais.available() == 1) {
			byte[] tmp = new byte[zdata2.length + 1];
			tmp[0] = zdata1[zdata1.length - 1];
			System.arraycopy(zdata2, 0, tmp, 1, zdata2.length);
			bais = new ByteArrayInputStream(tmp);
		} else {
			bais = new ByteArrayInputStream(zdata2);
		}
		zin.setRawStream(bais);
		assertTrue(Arrays.equals(Utils.readAll(zin), data2));
	}

	private void flushRand(int len)
		throws Exception
	{
		byte[] data1 = new byte[len];
		byte[] data2 = new byte[len];
		Utils.PRNG.nextBytes(data1);
		Utils.PRNG.nextBytes(data2);
		makeFlush(data1, data2, 0);
		makeFlush(data1, data2, 1);
		makeFlush(data1, data2, 2);
	}

	private void flushZero(int len)
		throws Exception
	{
		byte[] data1 = new byte[len];
		byte[] data2 = new byte[len];
		makeFlush(data1, data2, 0);
		makeFlush(data1, data2, 1);
		makeFlush(data1, data2, 2);
	}

	private void flushText(int len)
		throws Exception
	{
		byte[] data1 = Utils.makeRandText(len);
		byte[] data2 = Utils.makeRandText(len);
		makeFlush(data1, data2, 0);
		makeFlush(data1, data2, 1);
		makeFlush(data1, data2, 2);
	}

	public void testRand()
		throws Exception
	{
		for (int i = 2; i < 100; i ++)
			flushRand(i);
		int len = 100;
		for (int i = 0; i < 12; i ++) {
			len *= 2;
			flushRand(len);
		}
	}

	public void testZero()
		throws Exception
	{
		for (int i = 0; i < 100; i ++)
			flushZero(i);
		int len = 100;
		for (int i = 0; i < 12; i ++) {
			len *= 2;
			flushZero(len);
		}
	}

	public void testText()
		throws Exception
	{
		for (int i = 0; i < 100; i ++)
			flushText(i);
		int len = 100;
		for (int i = 0; i < 12; i ++) {
			len *= 2;
			flushText(len);
		}
	}
}
