// $Id$

package org.bolet.jgz.api;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import org.bolet.jgz.ZlibInputStream;
import org.bolet.jgz.ZlibOutputStream;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.io.IOException;
import java.util.Arrays;
import java.util.zip.DataFormatException;
import java.util.zip.Deflater;
import java.util.zip.DeflaterOutputStream;
import java.util.zip.Inflater;
import java.util.zip.InflaterInputStream;

/**
 * Unit tests for interoperability with zlib (standard JDK classes
 * from <code>java.util.zip</code>).
 *
 * @version   $Revision$
 * @author    Thomas Pornin
 */

public class TestInterop extends TestCase {

	public TestInterop(String name)
	{
		super(name);
	}

	public static Test suite()
	{
		return new TestSuite(TestInterop.class);
	}

	/* ============================================================ */

	private byte[] deflateRef(byte[] data, byte[] dict)
	{
		Deflater d = new Deflater();
		d.setDictionary(dict);
		d.setInput(data);
		d.finish();
		byte[] tmp = new byte[12 + data.length
			+ ((data.length + 99) / 100)];
		int rlen = d.deflate(tmp);
		byte[] res = new byte[rlen];
		System.arraycopy(tmp, 0, res, 0, rlen);
		return res;
	}

	private byte[] inflateRef(byte[] data, byte[] dict)
		throws DataFormatException
	{
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		byte[] buf = new byte[8192];
		Inflater i = new Inflater();
		i.setInput(data);
		boolean first = true;
		while (!i.finished()) {
			int len = i.inflate(buf);
			if (first) {
				if (len != 0)
					fail("did not request the dictionary");
				i.setDictionary(dict);
				first = false;
			} else {
				if (len > 0)
					baos.write(buf, 0, len);
			}
		}
		return baos.toByteArray();
	}

	private void interop(InputStream in)
		throws Exception
	{
		byte[] data = Utils.readAll(in);
		interop(data);
	}

	private void interop(byte[] data)
		throws Exception
	{
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		ZlibOutputStream zout = new ZlibOutputStream(baos);
		zout.write(data);
		zout.terminate();
		byte[] zdata1 = baos.toByteArray();
		baos = new ByteArrayOutputStream();
		DeflaterOutputStream dout = new DeflaterOutputStream(baos);
		dout.write(data);
		dout.close();
		byte[] zdata2 = baos.toByteArray();

		ByteArrayInputStream bais = new ByteArrayInputStream(zdata2);
		ZlibInputStream zin = new ZlibInputStream(bais);
		byte[] data2 = Utils.readAll(zin);
		assertTrue(Arrays.equals(data, data2));
		bais = new ByteArrayInputStream(zdata1);
		InflaterInputStream din = new InflaterInputStream(bais);
		byte[] data3 = Utils.readAll(din);
		assertTrue(Arrays.equals(data, data3));

		if (data.length == 0)
			return;

		int dictLen = data.length;
		if (dictLen > 32768)
			dictLen = 32768;
		byte[] dict = new byte[dictLen];
		System.arraycopy(data, data.length - dictLen, dict, 0, dictLen);

		baos = new ByteArrayOutputStream();
		zout = new ZlibOutputStream(baos);
		zout.processDictionary(dict);
		zout.write(data);
		zout.terminate();
		byte[] zdata3 = baos.toByteArray();
		byte[] zdata4 = deflateRef(data, dict);

		bais = new ByteArrayInputStream(zdata4);
		zin = new ZlibInputStream(bais);
		zin.processDictionary(dict);
		byte[] data4 = Utils.readAll(zin);
		assertTrue(Arrays.equals(data, data4));

		byte[] data5 = inflateRef(zdata3, dict);
		assertTrue(Arrays.equals(data, data5));

		/*
		if (data.length > 100)
			System.out.printf("%7d -> %6.2f%%\n",
				data.length, 100.0 *
				(double)zdata1.length / (double)zdata2.length);
		*/
	}

	private void interopRand(int len)
		throws Exception
	{
		byte[] data = new byte[len];
		Utils.PRNG.nextBytes(data);
		interop(data);
	}

	private void interopZero(int len)
		throws Exception
	{
		byte[] data = new byte[len];
		interop(data);
	}

	private void interopText(int len)
		throws Exception
	{
		byte[] data = Utils.makeRandText(len);
		interop(data);
	}

	public void testRand()
		throws Exception
	{
		for (int i = 0; i < 100; i ++)
			interopRand(i);
		int len = 100;
		for (int i = 0; i < 12; i ++) {
			len *= 2;
			interopRand(len);
		}
	}

	public void testZero()
		throws Exception
	{
		for (int i = 0; i < 100; i ++)
			interopZero(i);
		int len = 100;
		for (int i = 0; i < 12; i ++) {
			len *= 2;
			interopZero(len);
		}
	}

	public void testText()
		throws Exception
	{
		for (int i = 0; i < 100; i ++)
			interopText(i);
		int len = 100;
		for (int i = 0; i < 12; i ++) {
			len *= 2;
			interopText(len);
		}
	}
}
